/*******************************************************************************
 *
 * blueMarine - open source photo workflow
 * =======================================
 *
 * Copyright (C) 2003-2010 by Fabrizio Giudici
 * Project home page: http://bluemarine.tidalwave.it
 *
 *******************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *******************************************************************************
 *
 * $Id: VisualSemanticControllerImpl.java,v 8e2ec9f44519 2010/02/18 15:28:00 fabrizio $
 *
 ******************************************************************************/
package it.tidalwave.semantic.graph.impl;

import it.tidalwave.netbeans.visual.EdgeWidgetFactory;
import java.util.Collection;
import java.util.Set;
import javax.annotation.Nonnull;
import java.util.HashMap;
import java.util.Map;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.openide.nodemanager.NodeDelegateFactory;
import org.netbeans.api.visual.model.ObjectScene;
import org.openide.util.Parameters;
import it.tidalwave.semantic.other.Entity;
import it.tidalwave.semantic.other.Lookupable;
import it.tidalwave.semantic.other.Relation;
import it.tidalwave.semantic.other.Type;
import it.tidalwave.netbeans.visual.NodeScene;
import it.tidalwave.netbeans.visual.NodeWidgetFactory;
import it.tidalwave.semantic.graph.VisualSemanticController;
import it.tidalwave.semantic.graph.VisualSemanticController.EntityWidgetFactory;
import java.util.Collections;
import org.netbeans.api.visual.action.ActionFactory;
import org.netbeans.api.visual.anchor.AnchorShape;
import org.netbeans.api.visual.layout.LayoutFactory;
import org.netbeans.api.visual.widget.ConnectionWidget;
import org.netbeans.api.visual.widget.LabelWidget;
import org.netbeans.api.visual.widget.Widget;
import org.openide.nodes.Node;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: VisualSemanticControllerImpl.java,v 8e2ec9f44519 2010/02/18 15:28:00 fabrizio $
 *
 ******************************************************************************/
public class VisualSemanticControllerImpl implements VisualSemanticController
  {
    private final static String CLASS = VisualSemanticControllerImpl.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);

    @Nonnull
    private final NodeScene scene;

//    @Nonnull
//    private final WidgetAction moveAction = ActionFactory.createMoveAction();

    private final Map<Type, EntityWidgetFactory> entityWidgetFactoryMapByType = new HashMap<Type, EntityWidgetFactory>();

    private final Map<Type, RelationWidgetFactory> relationWidgetFactoryMapByType = new HashMap<Type, RelationWidgetFactory>();

    private final NamespaceManager namespaceManager = new NamespaceManager();

    /***************************************************************************
     *
     *
     **************************************************************************/
    final NodeWidgetFactory nodeWidgetFactory = new NodeWidgetFactory()
      {
        @Override
        public Collection<? extends Widget> createWidgets (@Nonnull final ObjectScene scene, @Nonnull final Node node)
          {
            final Widget widget = new DefaultEntityWidget(scene, node);
            widget.getActions().addAction(ActionFactory.createMoveAction());

            return Collections.singletonList(widget);
          }
      };

    /***************************************************************************
     *
     *
     **************************************************************************/
    final EdgeWidgetFactory edgeWidgetFactory = new EdgeWidgetFactory()
      {
        @Override
        public Collection<? extends ConnectionWidget> createConnectionWidget (@Nonnull final ObjectScene scene, @Nonnull final Node node)
          {
            final ConnectionWidget connectionWidget = new ConnectionWidget(scene);
            connectionWidget.setTargetAnchorShape(AnchorShape.TRIANGLE_FILLED);

            final LabelWidget connectionLabel = new LabelWidget(scene, namespaceManager.resolveURI(node.getDisplayName()));
            connectionLabel.setOpaque(true);
            connectionWidget.addChild(connectionLabel);
            connectionWidget.setConstraint(connectionLabel, LayoutFactory.ConnectionWidgetLayoutAlignment.CENTER, 0.5f);

            return Collections.singletonList(connectionWidget);
          }
      };



    /***************************************************************************
     *
     *
     **************************************************************************/
    public VisualSemanticControllerImpl (@Nonnull final NodeScene scene)
      {
        Parameters.notNull("scene", scene);
        this.scene = scene;
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void registerWidgetFactory (@Nonnull final Type type,
                                       @Nonnull final EntityWidgetFactory widgetFactory)
      {
        logger.fine("registerWidgetFactory(%s, %s)", type, widgetFactory);
        Parameters.notNull("type", type);
        Parameters.notNull("widgetFactory", widgetFactory);

        entityWidgetFactoryMapByType.put(type, widgetFactory);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void registerWidgetFactory (@Nonnull final Type type,
                                       @Nonnull final RelationWidgetFactory widgetFactory)
      {
        logger.fine("registerWidgetFactory(%s, %s)", type, widgetFactory);
        Parameters.notNull("type", type);
        Parameters.notNull("widgetFactory", widgetFactory);
        
        relationWidgetFactoryMapByType.put(type, widgetFactory);
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void addEntities (@Nonnull final Set<Entity> entities)
      {
        logger.fine("addEntities(%d entities)", entities.size());
        Parameters.notNull("entities", entities);

        for (final Entity entity : entities)
          {
            logger.finer(">>>> creating widget for Entity: %s", entity);

            final Node entityNode = findNode(entity);

            if (!scene.isNode(entityNode))
              {
                scene.addNode(entityNode);
              }

//            for (final Relation relation : entity.getRelations())
//              {
//                final Node subjectNode = findNode(relation.getSubject());
//                final Node objectNode = findNode(relation.getObject());
//
//                if (!scene.isNode(subjectNode))
//                  {
//                    scene.addNode(subjectNode);
//                  }
//
//                if (!scene.isNode(objectNode))
//                  {
//                    scene.addNode(objectNode);
//                  }
//              }
//          }
//
//        for (final Entity entity : entities)
//          {
            for (final Relation relation : entity.findRelations())
              {
                if (relation.getSubject() == entity)
                  {
                    final Node relationNode = findNode(relation);
                    final Node subjectNode = findNode(relation.getSubject());
                    final Node objectNode = findNode(relation.getObject());

                    if (!scene.isNode(subjectNode))
                      {
                        scene.addNode(subjectNode);
                      }

                    if (!scene.isNode(objectNode))
                      {
                        scene.addNode(objectNode);
                      }

                    scene.addEdge(relationNode);
                    scene.setEdgeSource(relationNode, subjectNode);
                    scene.setEdgeTarget(relationNode, objectNode);
                  }
              }
          }

        scene.validate();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
    public void clean()
      {
        scene.removeAllNodes();
      }

    /***************************************************************************
     *
     *
     **************************************************************************/
//    protected Widget createWidgetForEntity (@Nonnull final Node node)
//      {
//        final Entity entity = node.getLookup().lookup(Entity.class);
//        final EntityWidgetFactory widgetFactory = entityWidgetFactoryMapByType.get(entity.getType());
//
//        if (widgetFactory != null)
//          {
//            return widgetFactory.createWidget(scene, node);
//          }
//        else
//          {
//            final Widget widget = new DefaultEntityWidget(scene, node);
//            widget.getActions().addAction(moveAction);
//            widget.getActions().addAction(bringToFrontAction);
//            return widget;
//          }
//      }

    /***************************************************************************
     *
     *
     **************************************************************************/
//    protected ConnectionWidget createWidgetForRelation (@Nonnull final Node node)
//      {
//        final Relation relation = node.getLookup().lookup(Relation.class);
//        final RelationWidgetFactory widgetFactory = relationWidgetFactoryMapByType.get(relation.getType());
//
//        if (widgetFactory != null)
//          {
//            return widgetFactory.createWidget(scene, node);
//          }
//        else
//          {
//            final ConnectionWidget connectionWidget = new ConnectionWidget(scene);
//            connectionWidget.setTargetAnchorShape(AnchorShape.TRIANGLE_FILLED);
//            return connectionWidget;
//          }
//      }

    // FIXME: make it weak
    final Map<Object, Node> nodeFlyweightMap = new HashMap<Object, Node>();

    @Nonnull
    public Node findNode (@Nonnull final Object object)
      {
        Node node = nodeFlyweightMap.get(object);

        if (node == null)
          {
            final NodeDelegateFactory nodeFactory = ((Lookupable)object).getLookup().lookup(NodeDelegateFactory.class);
            assert nodeFactory != null : "No NodeFactory for " + object;
            final Object factory = (object instanceof Entity) ? nodeWidgetFactory : edgeWidgetFactory;
            node = new ViewerFilterNode(nodeFactory.findOrCreateNodeDelegate(), scene, factory);
            nodeFlyweightMap.put(object, node);
          }

        assert node != null;
        return node;
      }
  }
